import asyncio
from unittest import IsolatedAsyncioTestCase as TestCase
from unittest.mock import AsyncMock as CoroutineMock

from atlassian_jwt_auth.contrib.aiohttp import HTTPSPublicKeyRetriever, JWTAuthVerifier
from atlassian_jwt_auth.tests import test_verifier, utils


class SyncJWTAuthVerifier(JWTAuthVerifier):
    def __init__(self, *args, loop=None, **kwargs):
        if loop is None:
            loop = asyncio.get_event_loop()
        self.loop = loop
        super().__init__(*args, **kwargs)

    def verify_jwt(self, *args, **kwargs):
        return self.loop.run_until_complete(super().verify_jwt(*args, **kwargs))


class JWTAuthVerifierTestMixin(test_verifier.BaseJWTAuthVerifierTest):
    loop = None

    def _setup_mock_public_key_retriever(self, pub_key_pem):
        m_public_key_ret = CoroutineMock(spec=HTTPSPublicKeyRetriever)
        m_public_key_ret.retrieve.return_value = pub_key_pem.decode()
        return m_public_key_ret

    def _setup_jwt_auth_verifier(self, pub_key_pem, **kwargs):
        m_public_key_ret = self._setup_mock_public_key_retriever(pub_key_pem)
        return SyncJWTAuthVerifier(m_public_key_ret, loop=self.loop, **kwargs)


class JWTAuthVerifierRS256Test(
    utils.RS256KeyTestMixin, JWTAuthVerifierTestMixin, TestCase
):
    """Tests for aiohttp.JWTAuthVerifier class for RS256 algorithm"""


class JWTAuthVerifierES256Test(
    utils.ES256KeyTestMixin, JWTAuthVerifierTestMixin, TestCase
):
    """Tests for aiohttp.JWTAuthVerifier class for ES256 algorithm"""



import base64
import unittest

from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import serialization
from cryptography.hazmat.primitives.serialization import load_pem_private_key

from atlassian_jwt_auth.key import DataUriPrivateKeyRetriever
from atlassian_jwt_auth.signer import JWTAuthSigner
from atlassian_jwt_auth.tests import utils


def convert_key_pem_format_to_der_format(private_key_pem):
    private_key = load_pem_private_key(
        private_key_pem, password=None, backend=default_backend()
    )
    return private_key.private_bytes(
        encoding=serialization.Encoding.DER,
        format=serialization.PrivateFormat.PKCS8,
        encryption_algorithm=serialization.NoEncryption(),
    )


class BaseDataUriPrivateKeyRetrieverTest(object):
    """tests for the DataUriPrivateKeyRetriever class."""

    def setUp(self):
        self._private_key_pem = self.get_new_private_key_in_pem_format()
        self._public_key_pem = utils.get_public_key_pem_for_private_key_pem(
            self._private_key_pem
        )
        self._private_key_der = convert_key_pem_format_to_der_format(
            self._private_key_pem
        )

    def get_example_data_uri(self, private_key_der):
        return "data:application/pkcs8;kid=example%2Feg;base64," + base64.b64encode(
            private_key_der
        ).decode("utf-8")

    def test_load_data_uri(self):
        """tests that a valid data uri is correctly loaded."""
        expected_kid = "example/eg"
        data_uri = self.get_example_data_uri(self._private_key_der)
        provider = DataUriPrivateKeyRetriever(data_uri)
        kid, private_key_pem = provider.load("example")
        self.assertEqual(kid.key_id, expected_kid)
        self.assertEqual(private_key_pem, self._private_key_pem.decode("utf-8"))

    def test_load_data_uri_can_be_used_with_a_signer(self):
        """tests that the data uri private key retriever can be used with a
        signer to generate a jwt.
        """
        data_uri = self.get_example_data_uri(self._private_key_der)
        provider = DataUriPrivateKeyRetriever(data_uri)
        jwt_auth_signer = JWTAuthSigner("issuer", provider, algorithm=self.algorithm)
        jwt_auth_signer.generate_jwt("aud")


class DataUriPrivateKeyRetrieverRS256Test(
    BaseDataUriPrivateKeyRetrieverTest, utils.RS256KeyTestMixin, unittest.TestCase
):
    pass


class DataUriPrivateKeyRetrieverES256Test(
    BaseDataUriPrivateKeyRetrieverTest, utils.ES256KeyTestMixin, unittest.TestCase
):
    pass



import os
import shutil
import tempfile
import unittest

import atlassian_jwt_auth
from atlassian_jwt_auth import key
from atlassian_jwt_auth.tests import utils


class BaseJWTAuthSignerWithFilePrivateKeyRetrieverTest(object):
    """tests for the JWTAuthSigner using the FilePrivateKeyRetriever."""

    def setUp(self):
        self.test_dir = tempfile.mkdtemp(prefix="atlassian-jwt-p-tests")
        self.key_dir = os.path.join(self.test_dir, "jwtprivatekeys")
        for dir in ["invalid-issuer", "issuer-with-many-keys", "valid-issuer"]:
            os.makedirs(os.path.join(self.key_dir, dir))
        self._private_key_pem = self.get_new_private_key_in_pem_format()
        for file_loc in [
            "invalid-issuer/key-tests-pem.new",
            "issuer-with-many-keys/key1.pem.new",
            "issuer-with-many-keys/key2.pem",
            "issuer-with-many-keys/key3.pem",
            "issuer-with-many-keys/key4.pem.new",
            "valid-issuer/key-for-tests.pem",
        ]:
            file_location = os.path.join(self.key_dir, file_loc)
            with open(file_location, "wb") as f:
                f.write(self._private_key_pem)

    def tearDown(self):
        if self.test_dir:
            shutil.rmtree(self.test_dir)

    def create_signer_for_issuer(self, issuer):
        return atlassian_jwt_auth.create_signer_from_file_private_key_repository(
            issuer, self.key_dir, algorithm=self.algorithm
        )

    def test_succeeds_if_issuer_has_one_valid_key(self):
        signer = self.create_signer_for_issuer("valid-issuer")
        token = signer.generate_jwt("audience")
        self.assertIsNotNone(token)

    def test_picks_last_valid_key_id(self):
        signer = self.create_signer_for_issuer("issuer-with-many-keys")
        token = signer.generate_jwt("audience")
        key_identifier = key._get_key_id_from_jwt_header(token)
        expected_key_id = "issuer-with-many-keys/key3.pem"
        self.assertEqual(key_identifier.key_id, expected_key_id)

    def test_fails_if_issuer_has_no_valid_keys(self):
        signer = self.create_signer_for_issuer("invalid-issuer")
        with self.assertRaisesRegex(IOError, "Issuer has no valid keys"):
            signer.generate_jwt("audience")

    def test_fails_if_issuer_does_not_exist(self):
        signer = self.create_signer_for_issuer("this-does-not-exist")
        with self.assertRaisesRegex(OSError, "No such file or directory"):
            signer.generate_jwt("audience")


class JWTAuthSignerWithFilePrivateKeyRetrieverRS256Test(
    BaseJWTAuthSignerWithFilePrivateKeyRetrieverTest,
    utils.RS256KeyTestMixin,
    unittest.TestCase,
):
    pass


class JWTAuthSignerWithFilePrivateKeyRetrieverES256Test(
    BaseJWTAuthSignerWithFilePrivateKeyRetrieverTest,
    utils.ES256KeyTestMixin,
    unittest.TestCase,
):
    pass



import unittest
from datetime import timedelta
from typing import Any

import jwt
from requests import Request

import atlassian_jwt_auth
from atlassian_jwt_auth.contrib.requests import JWTAuth, create_jwt_auth
from atlassian_jwt_auth.tests import utils


class BaseRequestsTest(object):
    """tests for the contrib.requests.JWTAuth class"""

    auth_cls: Any = JWTAuth

    def setUp(self):
        self._private_key_pem = self.get_new_private_key_in_pem_format()
        self._public_key_pem = utils.get_public_key_pem_for_private_key_pem(
            self._private_key_pem
        )

    def assert_authorization_header_is_valid(self, auth):
        """asserts that the given request contains a valid Authorization
        header.
        """
        auth_header = self._get_auth_header(auth)
        bearer = auth_header.split(b" ")[1]
        # Decode the JWT (verifying the signature and aud match)
        # an exception is thrown if this fails
        algorithms = atlassian_jwt_auth.get_permitted_algorithm_names()
        return jwt.decode(
            bearer,
            self._public_key_pem.decode(),
            audience="audience",
            algorithms=algorithms,
        )

    def _get_auth_header(self, auth):
        request = auth(Request())
        auth_header = request.headers["Authorization"]
        return auth_header

    def create_jwt_auth(self, *args, **kwargs):
        return create_jwt_auth(*args, **kwargs)

    def test_JWTAuth_make_authenticated_request(self):
        """Verify a valid Authorization header is added by JWTAuth"""
        jwt_auth_signer = atlassian_jwt_auth.create_signer(
            "issuer",
            "issuer/key",
            self._private_key_pem.decode(),
            algorithm=self.algorithm,
        )
        auth = self.auth_cls(jwt_auth_signer, "audience")
        self.assert_authorization_header_is_valid(auth)

    def test_create_jwt_auth(self):
        """Verify a valid Authorization header is added by JWTAuth"""
        auth = self.create_jwt_auth(
            "issuer",
            "issuer/key",
            self._private_key_pem.decode(),
            "audience",
            algorithm=self.algorithm,
        )
        self.assert_authorization_header_is_valid(auth)

    def test_create_jwt_auth_with_additional_claims(self):
        """Verify a Valid Authorization header is added by JWTAuth and
        contains the additional claims when provided.
        """
        jwt_auth_signer = atlassian_jwt_auth.create_signer(
            "issuer",
            "issuer/key",
            self._private_key_pem.decode(),
            algorithm=self.algorithm,
        )
        auth = self.auth_cls(
            jwt_auth_signer, "audience", additional_claims={"example": "claim"}
        )
        token = self.assert_authorization_header_is_valid(auth)
        self.assertEqual(token.get("example"), "claim")

    def test_do_not_reuse_jwts(self):
        auth = self.create_jwt_auth(
            "issuer",
            "issuer/key",
            self._private_key_pem.decode(),
            "audience",
            algorithm=self.algorithm,
        )
        auth_header = self._get_auth_header(auth)
        self.assertNotEqual(auth_header, self._get_auth_header(auth))

    def test_reuse_jwts(self):
        auth = self.create_jwt_auth(
            "issuer",
            "issuer/key",
            self._private_key_pem.decode(),
            "audience",
            algorithm=self.algorithm,
            reuse_jwts=True,
        )
        auth_header = self._get_auth_header(auth)
        self.assertEqual(auth_header, self._get_auth_header(auth))

    def test_do_not_reuse_jwt_if_audience_changes(self):
        auth = self.create_jwt_auth(
            "issuer",
            "issuer/key",
            self._private_key_pem.decode(),
            "audience",
            algorithm=self.algorithm,
            reuse_jwts=True,
        )
        auth_header = self._get_auth_header(auth)
        auth._audience = "not-" + auth._audience
        self.assertNotEqual(auth_header, self._get_auth_header(auth))

    def test_do_not_reuse_jwt_if_issuer_changes(self):
        auth = self.create_jwt_auth(
            "issuer",
            "issuer/key",
            self._private_key_pem.decode(),
            "audience",
            algorithm=self.algorithm,
            reuse_jwts=True,
        )
        auth_header = self._get_auth_header(auth)
        auth._signer.issuer = "not-" + auth._signer.issuer
        self.assertNotEqual(auth_header, self._get_auth_header(auth))

    def test_do_not_reuse_jwt_if_lifetime_changes(self):
        auth = self.create_jwt_auth(
            "issuer",
            "issuer/key",
            self._private_key_pem.decode(),
            "audience",
            algorithm=self.algorithm,
            reuse_jwts=True,
        )
        auth_header = self._get_auth_header(auth)
        auth._signer.lifetime = auth._signer.lifetime - timedelta(seconds=1)
        self.assertNotEqual(auth_header, self._get_auth_header(auth))

    def test_do_not_reuse_jwt_if_subject_changes(self):
        auth = self.create_jwt_auth(
            "issuer",
            "issuer/key",
            self._private_key_pem.decode(),
            "audience",
            algorithm=self.algorithm,
            reuse_jwts=True,
            subject="subject",
        )
        auth_header = self._get_auth_header(auth)
        auth._signer.subject = "not-" + auth._signer.subject
        self.assertNotEqual(auth_header, self._get_auth_header(auth))

    def test_do_not_reuse_jwt_if_additional_claims_change(self):
        auth = self.create_jwt_auth(
            "issuer",
            "issuer/key",
            self._private_key_pem.decode(),
            "audience",
            algorithm=self.algorithm,
            reuse_jwts=True,
        )
        auth_header = self._get_auth_header(auth)
        auth._additional_claims["foo"] = "bar"
        self.assertNotEqual(auth_header, self._get_auth_header(auth))

    def test_reuse_jwt_with_additional_claims(self):
        # calculating the cache key with additional claims is non-trivial
        auth = self.create_jwt_auth(
            "issuer",
            "issuer/key",
            self._private_key_pem.decode(),
            "audience",
            algorithm=self.algorithm,
            reuse_jwts=True,
        )
        auth._additional_claims["foo"] = "bar"
        auth._additional_claims["fool"] = "blah"
        auth._additional_claims["foot"] = "quux"
        auth_header = self._get_auth_header(auth)
        self.assertEqual(auth_header, self._get_auth_header(auth))


class RequestsRS256Test(BaseRequestsTest, utils.RS256KeyTestMixin, unittest.TestCase):
    pass


class RequestsES256Test(BaseRequestsTest, utils.ES256KeyTestMixin, unittest.TestCase):
    pass



import unittest

from atlassian_jwt_auth.frameworks.common import utils


class SettingsDictTest(unittest.TestCase):
    """Tests for the SettingsDict class."""

    def test_hash(self) -> None:
        """Test that SettingsDict instances can be hashed."""
        dictionary_one = {"a": "b", "3": set([1]), "f": None}
        dictionary_two = {"a": "b", "3": set([1]), "f": None}
        dictionary_three = {"a": "b", "3": set([1]), "diff": "333"}
        settings_one = utils.SettingsDict(dictionary_one)
        settings_two = utils.SettingsDict(dictionary_two)
        settings_three = utils.SettingsDict(dictionary_three)
        self.assertEqual(settings_one, settings_two)
        self.assertEqual(hash(settings_one), hash(settings_two))
        self.assertNotEqual(settings_one, settings_three)
        self.assertNotEqual(hash(settings_one), hash(settings_three))



from .decorators import requires_asap, restrict_asap, with_asap
from .middleware import OldStyleASAPMiddleware, asap_middleware

__all__ = [
    "restrict_asap",
    "with_asap",
    "requires_asap",
    "OldStyleASAPMiddleware",
    "asap_middleware",
]



import datetime
import unittest
from unittest import mock

from cryptography.hazmat.primitives import serialization

import atlassian_jwt_auth
from atlassian_jwt_auth.tests import utils


class BaseJWTAuthSignerTest(object):
    """tests for the JWTAuthSigner class."""

    def setUp(self):
        self._private_key_pem = self.get_new_private_key_in_pem_format()

    def test__generate_claims(self):
        """tests that _generate_claims works as expected."""
        expected_now = datetime.datetime(year=2001, day=1, month=1)
        expected_audience = "example_aud"
        expected_iss = "eg"
        expected_key_id = "eg/ex"
        jwt_auth_signer = atlassian_jwt_auth.create_signer(
            expected_iss, expected_key_id, self._private_key_pem
        )
        jwt_auth_signer._now = lambda: expected_now
        for additional_claims in [{}, {"extra": "thing"}]:
            expected_claims = {
                "iss": expected_iss,
                "exp": expected_now + datetime.timedelta(minutes=1),
                "iat": expected_now,
                "aud": expected_audience,
                "nbf": expected_now,
                "sub": expected_iss,
            }
            expected_claims.update(additional_claims)
            claims = jwt_auth_signer._generate_claims(
                expected_audience, additional_claims=additional_claims
            )
            self.assertIsNotNone(claims["jti"])
            del claims["jti"]
            self.assertEqual(claims, expected_claims)

    def test_jti_changes(self):
        """tests that the jti of a claim changes."""
        expected_now = datetime.datetime(year=2001, day=1, month=1)
        aud = "aud"
        jwt_auth_signer = utils.get_example_jwt_auth_signer(
            algorithm=self.algorithm, private_key_pem=self._private_key_pem
        )
        jwt_auth_signer._now = lambda: expected_now
        first = jwt_auth_signer._generate_claims(aud)["jti"]
        second = jwt_auth_signer._generate_claims(aud)["jti"]
        self.assertNotEqual(first, second)
        self.assertTrue(str(expected_now.strftime("%s")) in first)
        self.assertTrue(str(expected_now.strftime("%s")) in second)

    @mock.patch("jwt.encode")
    def test_generate_jwt(self, m_jwt_encode):
        """tests that generate_jwt works as expected."""
        expected_aud = "aud_x"
        expected_claims = {"eg": "ex"}
        expected_key_id = "key_id"
        expected_issuer = "a_issuer"
        jwt_auth_signer = atlassian_jwt_auth.create_signer(
            expected_issuer,
            expected_key_id,
            private_key_pem=self._private_key_pem,
            algorithm=self.algorithm,
        )
        jwt_auth_signer._generate_claims = lambda aud: expected_claims
        jwt_auth_signer.generate_jwt(expected_aud)
        m_jwt_encode.assert_called_with(
            expected_claims,
            key=mock.ANY,
            algorithm=self.algorithm,
            headers={"kid": expected_key_id},
        )
        for name, args, kwargs in m_jwt_encode.mock_calls:
            if not kwargs:
                self.assertEqual(args[0], "utf-8")
                continue
            call_private_key = kwargs["key"].private_bytes(
                encoding=serialization.Encoding.PEM,
                format=serialization.PrivateFormat.TraditionalOpenSSL,
                encryption_algorithm=serialization.NoEncryption(),
            )
            self.assertEqual(call_private_key, self._private_key_pem)


class JWTAuthSignerRS256Test(
    BaseJWTAuthSignerTest, utils.RS256KeyTestMixin, unittest.TestCase
):
    pass


class JWTAuthSignerES256Test(
    BaseJWTAuthSignerTest, utils.ES256KeyTestMixin, unittest.TestCase
):
    pass



import logging
from typing import Any, Dict, Iterable, Optional

from jwt.exceptions import InvalidIssuerError, InvalidTokenError

from atlassian_jwt_auth import JWTAuthVerifier
from atlassian_jwt_auth.exceptions import (
    JtiUniquenessException,
    NoTokenProvidedError,
    PublicKeyRetrieverException,
    SubjectDoesNotMatchIssuerException,
)
from atlassian_jwt_auth.frameworks.common.backend import Backend
from atlassian_jwt_auth.frameworks.common.utils import SettingsDict


def _process_asap_token(
    request: Any,
    backend: Backend,
    settings: SettingsDict,
    verifier: Optional[JWTAuthVerifier] = None,
) -> Optional[str]:
    """Verifies an ASAP token, validates the claims, and returns an error
    response"""
    logger = logging.getLogger("asap")
    token = backend.get_asap_token(request)
    error_response = None
    if (
        token is None
        and not settings.ASAP_REQUIRED
        and (settings.ASAP_REQUIRED is not None)
    ):
        return None
    try:
        if token is None:
            raise NoTokenProvidedError
        if verifier is None:
            verifier = backend.get_verifier(settings=settings)
        asap_claims = verifier.verify_jwt(
            token,
            settings.ASAP_VALID_AUDIENCE,
            leeway=settings.ASAP_VALID_LEEWAY,
        )

        _verify_issuers(asap_claims, settings.ASAP_VALID_ISSUERS)
        backend.set_asap_claims_for_request(request, asap_claims)
    except NoTokenProvidedError:
        logger.info("No token provided")
        error_response = backend.get_401_response("Unauthorized", request=request)
    except PublicKeyRetrieverException as e:
        if e.status_code not in (403, 404):
            # Any error other than "not found" is a problem and should
            # be dealt with elsewhere.
            # Note that we treat 403 like 404 to account for the fact
            # that a server configured to secure directory listings
            # will return 403 for a missing file to avoid leaking
            # information.
            raise
        logger.warning("Could not retrieve the matching public key")
        error_response = backend.get_401_response(
            "Unauthorized: Key not found", request=request
        )
    except InvalidIssuerError:
        logger.warning("Invalid token - issuer")
        error_response = backend.get_403_response(
            "Forbidden: Invalid token issuer", request=request
        )
    except InvalidTokenError:
        logger.warning("Invalid token")
        error_response = backend.get_401_response(
            "Unauthorized: Invalid token", request=request
        )
    except JtiUniquenessException:
        logger.warning("Invalid token - duplicate jti")
        error_response = backend.get_401_response(
            "Unauthorized: Invalid token - duplicate jti", request=request
        )
    except SubjectDoesNotMatchIssuerException:
        logger.warning("Invalid token - subject and issuer do not match")
        error_response = backend.get_401_response(
            "Unauthorized: Subject and Issuer do not match", request=request
        )
    except Exception:
        logger.exception("An error occured while checking an asap token")
        raise

    if error_response is not None and settings.ASAP_REQUIRED:
        return error_response
    return None


def _verify_issuers(
    asap_claims: Dict[Any, Any], issuers: Optional[Iterable[str]] = None
) -> None:
    """Verify that the issuer in the claims is valid and is expected."""
    claim_iss = asap_claims.get("iss")
    if issuers is None:
        return None

    if claim_iss is None or claim_iss not in issuers:
        raise InvalidIssuerError
    return None



from __future__ import absolute_import

from typing import Any, Iterable, Union

import atlassian_jwt_auth
from atlassian_jwt_auth import KeyIdentifier
from atlassian_jwt_auth.signer import JWTAuthSigner


class BaseJWTAuth(object):
    """Adds a JWT bearer token to the request per the ASAP specification"""

    def __init__(
        self,
        signer: JWTAuthSigner,
        audience: Union[str, Iterable[str]],
        *args: Any,
        **kwargs: Any,
    ) -> None:
        self._audience = audience
        self._signer = signer
        self._additional_claims = kwargs.get("additional_claims", {})

    @classmethod
    def create(
        cls,
        issuer: str,
        key_identifier: Union[KeyIdentifier, str],
        private_key_pem: Union[str, bytes],
        audience: Union[str, Iterable[str]],
        **kwargs: Any,
    ) -> "BaseJWTAuth":
        """Instantiate a JWTAuth while creating the signer inline"""
        signer = atlassian_jwt_auth.create_signer(
            issuer, key_identifier, private_key_pem, **kwargs
        )
        return cls(signer, audience)

    def _get_header_value(self) -> bytes:
        return b"Bearer " + self._signer.generate_jwt(
            self._audience, additional_claims=self._additional_claims
        )



from __future__ import absolute_import

from typing import Any, Iterable, Union

import requests
from requests.auth import AuthBase

from atlassian_jwt_auth import KeyIdentifier
from atlassian_jwt_auth.auth import BaseJWTAuth


class JWTAuth(AuthBase, BaseJWTAuth):
    """Adds a JWT bearer token to the request per the ASAP specification"""

    def __call__(
        self, r: requests.models.PreparedRequest
    ) -> requests.models.PreparedRequest:
        r.headers["Authorization"] = self._get_header_value()  # type: ignore[assignment]
        return r


def create_jwt_auth(
    issuer: str,
    key_identifier: Union[KeyIdentifier, str],
    private_key_pem: str,
    audience: Union[str, Iterable[str]],
    **kwargs: Any,
) -> BaseJWTAuth:
    """Instantiate a JWTAuth while creating the signer inline"""
    return JWTAuth.create(issuer, key_identifier, private_key_pem, audience, **kwargs)



from typing import Any, Iterable, Union

from aiohttp import BasicAuth

from atlassian_jwt_auth import KeyIdentifier
from atlassian_jwt_auth.auth import BaseJWTAuth


class JWTAuth(BaseJWTAuth, BasicAuth):
    """Adds a JWT bearer token to the request per the ASAP specification

    It should be aiohttp.BasicAuth subclass, so redefine its `__new__` method.
    """

    def __new__(cls, *args, **kwargs):
        return super().__new__(cls, "")

    def encode(self) -> str:
        return self._get_header_value().decode(self.encoding)


def create_jwt_auth(
    issuer: str,
    key_identifier: Union[KeyIdentifier, str],
    private_key_pem: str,
    audience: Union[str, Iterable[str]],
    **kwargs: Any,
) -> BaseJWTAuth:
    """Instantiate a JWTAuth while creating the signer inline"""
    return JWTAuth.create(issuer, key_identifier, private_key_pem, audience, **kwargs)



from typing import Any, Dict, Type

import atlassian_jwt_auth
from atlassian_jwt_auth import JWTAuthVerifier
from atlassian_jwt_auth.key import BasePublicKeyRetriever


def get_static_retriever_class(keys: Dict[str, Any]) -> Type[BasePublicKeyRetriever]:
    class StaticPublicKeyRetriever(BasePublicKeyRetriever):
        """Retrieves a key from a static dict of public keys
        (for use in tests only)"""

        def __init__(self, *args: Any, **kwargs: Any) -> None:
            self.keys: Dict[str, Any] = keys

        def retrieve(self, key_identifier, **requests_kwargs) -> Any:
            return self.keys[key_identifier.key_id]

    return StaticPublicKeyRetriever


def static_verifier(keys: Dict[str, Any]) -> JWTAuthVerifier:
    return atlassian_jwt_auth.JWTAuthVerifier(get_static_retriever_class(keys)())



import datetime
import unittest
from unittest import mock

import jwt
import jwt.algorithms
import jwt.exceptions

import atlassian_jwt_auth
import atlassian_jwt_auth.exceptions
import atlassian_jwt_auth.key
import atlassian_jwt_auth.signer
from atlassian_jwt_auth.tests import utils


class NoneAlgorithmJwtAuthSigner(atlassian_jwt_auth.signer.JWTAuthSigner):
    """A JWTAuthSigner that generates JWTs using the none algorithm
    and supports specifying arbitrary alg jwt header values.
    """

    def generate_jwt(self, audience, **kwargs):
        alg_header = kwargs.get("alg_header", "none")
        key_identifier, private_key_pem = self.private_key_retriever.load(self.issuer)
        return jwt.encode(
            self._generate_claims(audience, **kwargs),
            algorithm=None,
            key=None,
            headers={"kid": key_identifier.key_id, "alg": alg_header},
        )


class BaseJWTAuthVerifierTest(object):
    """tests for the JWTAuthVerifier class."""

    def setUp(self):
        self._private_key_pem = self.get_new_private_key_in_pem_format()
        self._public_key_pem = utils.get_public_key_pem_for_private_key_pem(
            self._private_key_pem
        )
        self._example_aud = "aud_x"
        self._example_issuer = "egissuer"
        self._example_key_id = "%s/a" % self._example_issuer
        self._jwt_auth_signer = atlassian_jwt_auth.create_signer(
            self._example_issuer,
            self._example_key_id,
            self._private_key_pem.decode(),
            algorithm=self.algorithm,
        )

    def _setup_mock_public_key_retriever(self, pub_key_pem):
        m_public_key_ret = mock.Mock()
        m_public_key_ret.retrieve.return_value = pub_key_pem.decode()
        return m_public_key_ret

    def _setup_jwt_auth_verifier(self, pub_key_pem, **kwargs):
        m_public_key_ret = self._setup_mock_public_key_retriever(pub_key_pem)
        return atlassian_jwt_auth.JWTAuthVerifier(m_public_key_ret, **kwargs)

    def test_verify_jwt_with_valid_jwt(self):
        """test that verify_jwt verifies a valid jwt."""
        verifier = self._setup_jwt_auth_verifier(self._public_key_pem)
        signed_jwt = self._jwt_auth_signer.generate_jwt(self._example_aud)
        v_claims = verifier.verify_jwt(signed_jwt, self._example_aud)
        self.assertIsNotNone(v_claims)
        self.assertEqual(v_claims["aud"], self._example_aud)
        self.assertEqual(v_claims["iss"], self._example_issuer)

    def test_verify_jwt_with_none_algorithm(self):
        """tests that verify_jwt does not accept jwt that use the none
        algorithm.
        """
        verifier = self._setup_jwt_auth_verifier(self._public_key_pem)
        private_key_ret = atlassian_jwt_auth.key.StaticPrivateKeyRetriever(
            self._example_key_id, self._private_key_pem.decode()
        )
        jwt_signer = NoneAlgorithmJwtAuthSigner(
            issuer=self._example_issuer,
            private_key_retriever=private_key_ret,
        )
        for algorithm in ["none", "None", "nOne", "nonE", "NONE"]:
            if algorithm != "none":
                jwt.register_algorithm(algorithm, jwt.algorithms.NoneAlgorithm())
            jwt_token = jwt_signer.generate_jwt(self._example_aud, alg_header=algorithm)
            if algorithm != "none":
                jwt.unregister_algorithm(algorithm)
            jwt_headers = jwt.get_unverified_header(jwt_token)
            self.assertEqual(jwt_headers["alg"], algorithm)
            with self.assertRaises(jwt.exceptions.InvalidAlgorithmError):
                verifier.verify_jwt(jwt_token, self._example_aud)

    def test_verify_jwt_with_key_identifier_not_starting_with_issuer(self):
        """tests that verify_jwt rejects a jwt if the key identifier does
        not start with the claimed issuer.
        """
        verifier = self._setup_jwt_auth_verifier(self._public_key_pem)
        signer = atlassian_jwt_auth.create_signer(
            "issuer",
            "issuerx",
            self._private_key_pem.decode(),
            algorithm=self.algorithm,
        )
        a_jwt = signer.generate_jwt(self._example_aud)
        with self.assertRaisesRegex(ValueError, "Issuer does not own"):
            verifier.verify_jwt(a_jwt, self._example_aud)

    @mock.patch("atlassian_jwt_auth.verifier.jwt.decode")
    def test_verify_jwt_with_non_matching_sub_and_iss(self, m_j_decode):
        """tests that verify_jwt rejects a jwt if the claims
        contains a subject which does not match the issuer.
        """
        expected_msg = "Issuer does not match the subject"
        m_j_decode.return_value = {
            "iss": self._example_issuer,
            "sub": self._example_issuer[::-1],
        }
        a_jwt = self._jwt_auth_signer.generate_jwt(self._example_aud)
        verifier = self._setup_jwt_auth_verifier(self._public_key_pem)
        for exception in [
            ValueError,
            atlassian_jwt_auth.exceptions.SubjectDoesNotMatchIssuerException,
        ]:
            with self.assertRaisesRegex(exception, expected_msg):
                verifier.verify_jwt(a_jwt, self._example_aud)

    @mock.patch("atlassian_jwt_auth.verifier.jwt.decode")
    def test_verify_jwt_with_jwt_lasting_gt_max_time(self, m_j_decode):
        """tests that verify_jwt rejects a jwt if the claims
        period of validity is greater than the allowed maximum.
        """
        expected_msg = "exceeds the maximum"
        claims = self._jwt_auth_signer._generate_claims(self._example_aud)
        claims["iat"] = claims["exp"] - datetime.timedelta(minutes=61)
        for key in ["iat", "exp"]:
            claims[key] = claims[key].strftime("%s")
        m_j_decode.return_value = claims
        a_jwt = self._jwt_auth_signer.generate_jwt(self._example_aud)
        verifier = self._setup_jwt_auth_verifier(self._public_key_pem)
        with self.assertRaisesRegex(ValueError, expected_msg):
            verifier.verify_jwt(a_jwt, self._example_aud)

    def test_verify_jwt_with_jwt_with_already_seen_jti(self):
        """tests that verify_jwt rejects a jwt if the jti
        has already been seen.
        """
        verifier = self._setup_jwt_auth_verifier(
            self._public_key_pem, check_jti_uniqueness=True
        )
        a_jwt = self._jwt_auth_signer.generate_jwt(self._example_aud)
        self.assertIsNotNone(verifier.verify_jwt(a_jwt, self._example_aud))
        for exception in [
            ValueError,
            atlassian_jwt_auth.exceptions.JtiUniquenessException,
        ]:
            with self.assertRaisesRegex(exception, "has already been used"):
                verifier.verify_jwt(a_jwt, self._example_aud)

    def assert_jwt_accepted_more_than_once(self, verifier, a_jwt):
        """asserts that the given jwt is accepted more than once."""
        for i in range(0, 3):
            self.assertIsNotNone(verifier.verify_jwt(a_jwt, self._example_aud))

    def test_verify_jwt_with_already_seen_jti_with_uniqueness_disabled(self):
        """tests that verify_jwt accepts a jwt if the jti
        has already been seen and the verifier has been set
        to not check the uniqueness of jti.
        """
        verifier = self._setup_jwt_auth_verifier(
            self._public_key_pem, check_jti_uniqueness=False
        )
        a_jwt = self._jwt_auth_signer.generate_jwt(self._example_aud)
        self.assert_jwt_accepted_more_than_once(verifier, a_jwt)

    def test_verify_jwt_with_already_seen_jti_default(self):
        """tests that verify_jwt by default accepts a jwt if the jti
        has already been seen.
        """
        verifier = self._setup_jwt_auth_verifier(self._public_key_pem)
        a_jwt = self._jwt_auth_signer.generate_jwt(self._example_aud)
        self.assert_jwt_accepted_more_than_once(verifier, a_jwt)

    def test_verify_jwt_subject_should_match_issuer(self):
        verifier = self._setup_jwt_auth_verifier(
            self._public_key_pem, subject_should_match_issuer=True
        )
        a_jwt = self._jwt_auth_signer.generate_jwt(
            self._example_aud, additional_claims={"sub": "not-" + self._example_issuer}
        )
        with self.assertRaisesRegex(ValueError, "Issuer does not match the subject."):
            verifier.verify_jwt(a_jwt, self._example_aud)

    def test_verify_jwt_subject_does_not_need_to_match_issuer(self):
        verifier = self._setup_jwt_auth_verifier(
            self._public_key_pem, subject_should_match_issuer=False
        )
        a_jwt = self._jwt_auth_signer.generate_jwt(
            self._example_aud, additional_claims={"sub": "not-" + self._example_issuer}
        )
        self.assertIsNotNone(verifier.verify_jwt(a_jwt, self._example_aud))

    @mock.patch("atlassian_jwt_auth.verifier.jwt.decode")
    def test_verify_jwt_with_missing_aud_claim(self, m_j_decode):
        """tests that verify_jwt rejects jwt that do not have an aud
        claim.
        """
        expected_msg = (
            "Claims validity, the aud claim must be provided and cannot be empty."
        )
        claims = self._jwt_auth_signer._generate_claims(self._example_aud)
        del claims["aud"]
        m_j_decode.return_value = claims
        a_jwt = self._jwt_auth_signer.generate_jwt(self._example_aud)
        verifier = self._setup_jwt_auth_verifier(self._public_key_pem)
        with self.assertRaisesRegex(KeyError, expected_msg):
            verifier.verify_jwt(a_jwt, self._example_aud)

    def test_verify_jwt_with_none_aud(self):
        """tests that verify_jwt rejects jwt that have a None aud claim."""
        verifier = self._setup_jwt_auth_verifier(self._public_key_pem)
        a_jwt = self._jwt_auth_signer.generate_jwt(
            self._example_aud, additional_claims={"aud": None}
        )
        exceptions = (
            jwt.exceptions.InvalidAudienceError,
            jwt.exceptions.InvalidTokenError,
        )
        with self.assertRaises(exceptions) as cm:
            verifier.verify_jwt(a_jwt, self._example_aud)
        if not isinstance(cm.exception, jwt.exceptions.InvalidAudienceError):
            self.assertIn("aud", str(cm.exception))

    def test_verify_jwt_with_non_matching_aud(self):
        """tests that verify_jwt rejects a jwt if the aud claim does not
        match the given & expected audience.
        """
        verifier = self._setup_jwt_auth_verifier(self._public_key_pem)
        a_jwt = self._jwt_auth_signer.generate_jwt(
            self._example_aud,
            additional_claims={"aud": self._example_aud + "-different"},
        )
        with self.assertRaises(jwt.exceptions.InvalidAudienceError):
            verifier.verify_jwt(a_jwt, self._example_aud)


class JWTAuthVerifierRS256Test(
    BaseJWTAuthVerifierTest, utils.RS256KeyTestMixin, unittest.TestCase
):
    pass


class JWTAuthVerifierES256Test(
    BaseJWTAuthVerifierTest, utils.ES256KeyTestMixin, unittest.TestCase
):
    pass



from typing import Callable, Iterable, Optional

from ..common.backend import Backend
from ..common.decorators import _restrict_asap, _with_asap
from .backend import DjangoBackend


def with_asap(
    func: Optional[Callable] = None,
    issuers: Optional[Iterable[str]] = None,
    required: bool = True,
    subject_should_match_issuer: Optional[bool] = None,
) -> Callable:
    """Decorator to allow endpoint-specific ASAP authentication.

    If authentication fails, a 401 or 403 response will be returned. Otherwise,
    the decorated function will be executed.

    The ASAP claimset will be set on request.asap_claims for further
    inspection later in the request lifecycle.

    :param list func: The view to decorate.
    :param list issuers: A list of valid token issuers that can access this
                         endpoint.
    :param boolean required: Whether or not to require ASAP on this endpoint.
    :param boolean subject_should_match_issuer: Indicate whether the subject
                                                must match the issuer for a
                                                token to be considered valid.
    """
    return _with_asap(
        func, DjangoBackend(), issuers, required, subject_should_match_issuer
    )


def restrict_asap(
    func: Optional[Callable] = None,
    backend: Optional[Backend] = None,
    issuers: Optional[Iterable[str]] = None,
    required: bool = True,
    subject_should_match_issuer: Optional[bool] = None,
) -> Callable:
    """Decorator to allow endpoint-specific ASAP authorization policies.

    This decorator assumes that request.asap_claims has previously been set by
    the asap_middleware.

    If the token does not meet the requirements imposed by the decorator, a 401
    or 403 response will be returned. Otherwise, the decorated function will be
    executed.

    :param list func: The view to decorate.
    :param list issuers: A list of valid token issuers that can access this
                         endpoint.
    :param boolean required: Whether or not to require ASAP on this endpoint.
    :param boolean subject_should_match_issuer: Indicate whether the subject
                                                must match the issuer for a
                                                token to be considered valid.
    """
    return _restrict_asap(
        func,
        DjangoBackend(),
        issuers,
        required,
        subject_should_match_issuer=None,
    )


def requires_asap(
    issuers: Optional[Iterable[str]] = None,
    subject_should_match_issuer: Optional[bool] = None,
    func: Optional[Callable] = None,
) -> Callable:
    """Decorator for Django endpoints to require ASAP

    :param list issuers: *required The 'iss' claims that this endpoint is from.
    """
    return with_asap(
        func=func,
        required=True,
        issuers=issuers,
        subject_should_match_issuer=subject_should_match_issuer,
    )



from collections import OrderedDict
from functools import lru_cache
from typing import Any, Dict, Iterable, Optional, Sequence, Union

import jwt
import jwt.api_jwt
from cryptography.hazmat.primitives.asymmetric.ec import EllipticCurvePublicKey
from cryptography.hazmat.primitives.asymmetric.ed448 import Ed448PublicKey
from cryptography.hazmat.primitives.asymmetric.ed25519 import Ed25519PublicKey
from cryptography.hazmat.primitives.asymmetric.rsa import RSAPublicKey
from jwt import PyJWK
from jwt.exceptions import InvalidAlgorithmError
from jwt.types import Options

from atlassian_jwt_auth import KeyIdentifier, algorithms, exceptions, key
from atlassian_jwt_auth.key import BasePublicKeyRetriever

AllowedPublicKeys = Union[
    RSAPublicKey, EllipticCurvePublicKey, Ed25519PublicKey, Ed448PublicKey
]


@lru_cache(maxsize=10)
def _load_public_key(
    algorithms: Sequence[str], public_key: str, algorithm: Optional[str]
) -> Any:
    """Returns a public key object instance given the public key and
    algorithm.

    This has been extracted out of JWTAuthVerifier to avoid possible memory
    leaks via retained instance references.
    """
    if isinstance(public_key, (RSAPublicKey, EllipticCurvePublicKey)):
        return public_key
    if algorithm is None or algorithm not in algorithms:
        raise InvalidAlgorithmError("The specified alg value is not allowed")
    py_jws = jwt.api_jws.PyJWS(algorithms=algorithms)
    alg_obj = py_jws._algorithms[algorithm]
    return alg_obj.prepare_key(public_key)


class JWTAuthVerifier(object):
    """This class can be used to verify a JWT."""

    def __init__(
        self, public_key_retriever: BasePublicKeyRetriever, **kwargs: Any
    ) -> None:
        self.public_key_retriever = public_key_retriever
        self.algorithms = algorithms.get_permitted_algorithm_names()
        self._seen_jti: OrderedDict[str, None] = OrderedDict()
        self._subject_should_match_issuer = kwargs.get(
            "subject_should_match_issuer", True
        )
        self._check_jti_uniqueness = kwargs.get("check_jti_uniqueness", False)

    def verify_jwt(
        self,
        a_jwt: Union[str, bytes],
        audience: str,
        leeway: int = 0,
        **requests_kwargs: Any,
    ) -> Dict[Any, Any]:
        """Verify if the token is correct

        Returns:
             dict: the claims of the given jwt if verification is successful.

        Raises:
            ValueError: if verification failed.
        """
        key_identifier = key._get_key_id_from_jwt_header(a_jwt)
        public_key = self._retrieve_pub_key(key_identifier, requests_kwargs)

        alg = jwt.get_unverified_header(a_jwt).get("alg", None)
        public_key_obj = self._load_public_key(public_key, alg or "RS256")
        return self._decode_jwt(
            a_jwt, key_identifier, public_key_obj, audience=audience, leeway=leeway
        )

    def _retrieve_pub_key(
        self, key_identifier: Union[KeyIdentifier, str], requests_kwargs: Any
    ) -> str:
        return self.public_key_retriever.retrieve(key_identifier, **requests_kwargs)

    def _load_public_key(self, public_key: str, algorithm: Optional[str]) -> Any:
        """Returns a public key object instance given the public key and
        algorithm.
        """
        return _load_public_key(tuple(self.algorithms), public_key, algorithm)

    def _decode_jwt(
        self,
        a_jwt: Union[str, bytes],
        key_identifier: KeyIdentifier,
        jwt_key: Union[AllowedPublicKeys, PyJWK, str, bytes],
        audience: Optional[Union[str, Iterable[str]]] = None,
        leeway: int = 0,
    ) -> Dict[Any, Any]:
        """Decode JWT and check if it's valid"""
        options: Options = Options(
            verify_signature=True,
            require=["exp", "iat"],
            verify_exp=True,
            verify_iat=True,
        )

        claims = jwt.decode(
            a_jwt,
            key=jwt_key,
            algorithms=self.algorithms,
            options=options,
            audience=audience,
            leeway=leeway,
        )

        if (
            not key_identifier.key_id.startswith("%s/" % claims["iss"])
            and key_identifier.key_id != claims["iss"]
        ):
            raise ValueError("Issuer does not own the supplied public key")

        if self._subject_should_match_issuer and (
            claims.get("sub") and claims["iss"] != claims["sub"]
        ):
            raise exceptions.SubjectDoesNotMatchIssuerException(
                "Issuer does not match the subject."
            )

        _aud = claims.get("aud", None)
        if _aud is None:
            _msg = (
                "Claims validity, the aud claim must be provided and cannot be empty."
            )
            raise KeyError(_msg)
        _exp = int(claims["exp"])
        _iat = int(claims["iat"])
        if _exp - _iat > 3600:
            _msg = "Claims validity, '%s', exceeds the maximum 1 hour." % (_exp - _iat)
            raise ValueError(_msg)
        _jti = claims["jti"]
        if self._check_jti_uniqueness:
            self._check_jti(_jti)
        return claims

    def _check_jti(self, jti: str) -> None:
        """Checks that the given jti has not been already been used."""
        if jti in self._seen_jti:
            raise exceptions.JtiUniquenessException(
                "The jti, '%s', has already been used." % jti
            )
        self._seen_jti[jti] = None
        while len(self._seen_jti) > 1000:
            self._seen_jti.popitem(last=False)
